function obj = learn( obj, iternum, code, labels, codebook, mu, sigma, cache_path )
% Perform learning
%
% Copyright (c) 2014, Yoshitaka Ushiku
% This software is released under the MIT License.
% http://opensource.org/licenses/mit-license.php

if nargin < 5 || isempty(codebook)
    codebook = [];
end
if nargin < 6 || isempty(mu)
    obj.mu = [];
end
if nargin < 7 || isempty(sigma)
    sigma = [];
    obj.sigma = [];
end
if nargin < 8
    iscache = false;
else
    iscache = true;
end

type = obj.type;
if iscache
    for ind=1:iternum
        cur_cache_path = [cache_path '_iter' num2str(ind) '.mat'];
        if exist(cur_cache_path,'file')
            disp(['ALREADY EXISTS: ' cur_cache_path])
            load(cur_cache_path)
            continue;
        end
        if ind < iternum
            switch lower(type)
                case lower('PerceptronuOVR')
                    mu = PerceptronuOVR(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PerceptronwOVR')
                    mu = PerceptronwOVR(obj.ol_param.lambda, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PerceptronMUL')
                    mu = PerceptronMUL(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PAuOVR')
                    mu = PAuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu);
                case lower('PAwOVR')
                    mu = PAwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu);
                case lower('PA')
                    mu = PA(obj.ol_param.C, 1, code, labels, codebook, obj.mu);
                case lower('CWuOVR')
                    [mu, sigma] = CWuOVR(obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('CWwOVR')
                    [mu, sigma] = CWwOVR(obj.ol_param.eta, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('CW')
                    [mu, sigma] = CW(obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROWuOVR')
                    [mu, sigma] = AROWuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROWwOVR')
                    [mu, sigma] = AROWwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROW')
                    [mu, sigma] = AROW(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERDuOVR')
                    [mu, sigma] = NHERDuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERDwOVR')
                    [mu, sigma] = NHERDwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERD')
                    [mu, sigma] = NHERD(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCWuOVR')
                    [mu, sigma] = SCWuOVR(obj.ol_param.C, obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCWwOVR')
                    [mu, sigma] = SCWwOVR(obj.ol_param.C, obj.ol_param.eta, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCW')
                    [mu, sigma] = SCW(obj.ol_param.C, obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SGDSVMuOVR')
                    mu = SGDSVMuOVR(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('SGDSVMwOVR')
                    mu = SGDSVMwOVR(obj.ol_param.lambda, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('SGDSVMMUL')
                    mu = SGDSVMMUL(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
            end
            if exist('mu',      'var'); obj.mu      = mu;       end
            if exist('sigma',   'var'); obj.sigma   = sigma;    end
        else
            switch lower(type)
                case lower('PerceptronuOVR')
                    [avgmu, mu] = PerceptronuOVR(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PerceptronwOVR')
                    [avgmu, mu] = PerceptronwOVR(obj.ol_param.lambda, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PerceptronMUL')
                    [avgmu, mu] = PerceptronMUL(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('PAuOVR')
                    [avgmu, mu] = PAuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu);
                case lower('PAwOVR')
                    [avgmu, mu] = PAwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu);
                case lower('PA')
                    [avgmu, mu] = PA(obj.ol_param.C, 1, code, labels, codebook, obj.mu);
                case lower('CWuOVR')
                    [avgmu, mu, sigma] = CWuOVR(obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('CWwOVR')
                    [avgmu, mu, sigma] = CWwOVR(obj.ol_param.eta, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('CW')
                    [avgmu, mu, sigma] = CW(obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROWuOVR')
                    [avgmu, mu, sigma] = AROWuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROWwOVR')
                    [avgmu, mu, sigma] = AROWwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('AROW')
                    [avgmu, mu, sigma] = AROW(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERDuOVR')
                    [avgmu, mu, sigma] = NHERDuOVR(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERDwOVR')
                    [avgmu, mu, sigma] = NHERDwOVR(obj.ol_param.C, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('NHERD')
                    [avgmu, mu, sigma] = NHERD(obj.ol_param.C, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCWuOVR')
                    [avgmu, mu, sigma] = SCWuOVR(obj.ol_param.C, obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCWwOVR')
                    [avgmu, mu, sigma] = SCWwOVR(obj.ol_param.C, obj.ol_param.eta, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SCW')
                    [avgmu, mu, sigma] = SCW(obj.ol_param.C, obj.ol_param.eta, 1, code, labels, codebook, obj.mu, obj.sigma);
                case lower('SGDSVMuOVR')
                    [avgmu, mu] = SGDSVMuOVR(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('SGDSVMwOVR')
                    [avgmu, mu] = SGDSVMwOVR(obj.ol_param.lambda, obj.ol_param.beta, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
                case lower('SGDSVMMUL')
                    [avgmu, mu] = SGDSVMMUL(obj.ol_param.lambda, 1, code, labels, codebook, obj.mu, obj.ol_param.t_zero+(ind-1)*length(labels));
            end
            if exist('avgmu',   'var'); obj.avgmu   = avgmu;    end
            if exist('mu',      'var'); obj.mu      = mu;       end
            if exist('sigma',   'var'); obj.sigma   = sigma;    end
        end
        save(cur_cache_path,'obj','-v7.3')
    end
else
    switch lower(type)
        case lower('PerceptronuOVR')
            [avgmu, mu] = PerceptronuOVR(obj.ol_param.lambda, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
        case lower('PerceptronwOVR')
            [avgmu, mu] = PerceptronwOVR(obj.ol_param.lambda, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
        case lower('PerceptronMUL')
            [avgmu, mu] = PerceptronMUL(obj.ol_param.lambda, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
        case lower('PAuOVR')
            [avgmu, mu] = PAuOVR(obj.ol_param.C, iternum, code, labels, codebook, obj.mu);
        case lower('PAwOVR')
            [avgmu, mu] = PAwOVR(obj.ol_param.C, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu);
        case lower('PA')
            [avgmu, mu] = PA(obj.ol_param.C, iternum, code, labels, codebook, obj.mu);
        case lower('CWuOVR')
            [avgmu, mu, sigma] = CWuOVR(obj.ol_param.eta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('CWwOVR')
            [avgmu, mu, sigma] = CWwOVR(obj.ol_param.eta, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('CW')
            [avgmu, mu, sigma] = CW(obj.ol_param.eta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('AROWuOVR')
            [avgmu, mu, sigma] = AROWuOVR(obj.ol_param.C, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('AROWwOVR')
            [avgmu, mu, sigma] = AROWwOVR(obj.ol_param.C, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('AROW')
            [avgmu, mu, sigma] = AROW(obj.ol_param.C, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('NHERDuOVR')
            [avgmu, mu, sigma] = NHERDuOVR(obj.ol_param.C, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('NHERDwOVR')
            [avgmu, mu, sigma] = NHERDwOVR(obj.ol_param.C, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('NHERD')
            [avgmu, mu, sigma] = NHERD(obj.ol_param.C, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('SCWuOVR')
            [avgmu, mu, sigma] = SCWuOVR(obj.ol_param.C, obj.ol_param.eta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('SCWwOVR')
            [avgmu, mu, sigma] = SCWwOVR(obj.ol_param.C, obj.ol_param.eta, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('SCW')
            [avgmu, mu, sigma] = SCW(obj.ol_param.C, obj.ol_param.eta, iternum, code, labels, codebook, obj.mu, obj.sigma);
        case lower('SGDSVMuOVR')
            [avgmu, mu] = SGDSVMuOVR(obj.ol_param.lambda, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
        case lower('SGDSVMwOVR')
            [avgmu, mu] = SGDSVMwOVR(obj.ol_param.lambda, obj.ol_param.beta, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
        case lower('SGDSVMMUL')
            [avgmu, mu] = SGDSVMMUL(obj.ol_param.lambda, iternum, code, labels, codebook, obj.mu, obj.ol_param.t_zero);
    end
    if exist('avgmu',   'var'); obj.avgmu   = avgmu;    end
    if exist('mu',      'var'); obj.mu      = mu;       end
    if exist('sigma',   'var'); obj.sigma   = sigma;    end
end

end
