function C = pq_kmeans(X, k, G, iter_lim)
%
% Copyright (c) 2014, Yoshitaka Ushiku
% This software is released under the MIT License.
% http://opensource.org/licenses/mit-license.php

if nargin < 4
    iter_lim = 100;
end
if nargin < 3
    G = 8;
end
if nargin < 2
    k = 256;
end

% % toy data
% G = 2;
% k = 3;
% X = rand(8,15);
% iter_lim = 100;

% data info
dim = size(X,1);
if mod(dim,G) ~= 0
    error('illegal dimension')
end
up_num = dim/G;
num = size(X,2);

% make C a sparse matrix
C = sparse(1:dim,repmat(1:k:(up_num-1)*k+1,G,1),X(:,1),dim,k*up_num);
new_C = C(:,1:k:k*(up_num-1)+1);
max_D = -Inf(1,num*up_num);

% Distance
sv_norms = sum(reshape(X,G,num*up_num).^2,1);
for seedind = 2:k
    fprintf('initial center: %d\r',seedind);
    cur_rand = rand(1); 
    % L2 distance
    cur_D = reshape(bsxfun(@minus,2*X'*new_C,dot(new_C,new_C))',1,num*up_num);
    % minimum distance per datum
    max_D = max(cur_D,max_D);
    D = reshape(max_D+sv_norms,up_num,num);
    % make each sum 1
    D = bsxfun(@rdivide,cumsum(D,2),sum(D,2));
    % find new center
    new_c_ind = D >cur_rand & D ~= 0;
    new_c_ind(:,2:end) = new_c_ind(:,1:end-1) - new_c_ind(:,2:end);
    new_c_ind = new_c_ind(:)';
    new_c_ind = reshape(new_c_ind(ones(1,G),:),dim,num);
    % add new centers
    new_c_ind_y = repmat(seedind+(0:k:k*up_num-1),G,1);
    new_c_ind_x = 1:dim;
    C = C+sparse(new_c_ind_x,new_c_ind_y,sum(X.*new_c_ind,2),dim,k*up_num);
    % memorize for next seed
    new_C = C(:,seedind:k:k*(up_num-1)+seedind);
end

% prepare for parallel kmeans
iter = 1;
last = zeros(1,up_num*num);
C_ind_x = reshape(repmat(reshape(1:dim,G,up_num),k,1),1,dim*k);
C_ind_y = reshape(repmat(1:k*up_num,G,1),1,dim*k);
label_addition = repmat(0:k:k*(up_num-1),1,num);
cflag = true(1,up_num);
cur_up_num = sum(cflag);
% flag per dimension of all labels
lflag = repmat(cflag,1,num);
% flag per dimension of X
xflag = reshape(cflag(ones(G,1),:),  1,dim);
% flag per dimension of centers
cflag = reshape(cflag(ones(G*k,1),:),1,dim*k);

% assign samples to the nearest centers
[~,label] = max(reshape(bsxfun(@minus,2*X'*C,dot(C,C))',k,num*up_num),[],1);
label = label + repmat(0:k:k*(up_num-1),1,num);
C(:) = 0;

while iter <= iter_lim && any(label ~= last)
    fprintf('kmeans iter: %d (%d/%d changed their cluster, %d/%d finished.)\r',iter,sum(label~=last),num*up_num,up_num-cur_up_num,up_num);

    % transform label into indicator matrix
    E = sparse(1:cur_up_num*num,label(lflag),1,cur_up_num*num,up_num*k);
    % compute center of each cluster
    cur_C = reshape(X(xflag,:),G,cur_up_num*num)*(E*spdiags(1./sum(E,1)',0,up_num*k,up_num*k));
    temp_C = sparse(C_ind_x(cflag), C_ind_y(cflag), cur_C(cflag), dim, k*up_num);
    % memorize the last labels
    last = label;
    last_cflag = cflag;
    % assign samples to the new centers
    [~,part_label] = max(reshape(bsxfun(@minus,2*X(xflag,:)'*temp_C(xflag,:),dot(temp_C(xflag,:),temp_C(xflag,:)))',k,num*up_num),[],1);
    label(lflag) = part_label(lflag) + label_addition(lflag);

    % find changed unit
    cflag = histc(last(last~=label),(0:k:k*up_num)+.5) ~= 0;
    cflag = cflag(1:end-1);
    cur_up_num = sum(cflag);
    % flag per dimension of all labels
    lflag = repmat(cflag,1,num);
    % flag per dimension of X
    xflag = reshape(cflag(ones(G,1),:),  1,dim);
    % flag per dimension of centers
    cflag = reshape(cflag(ones(G*k,1),:),1,dim*k);

    % update final C
    new_center = xor(cflag,last_cflag);
    C = C+sparse(C_ind_x(new_center), C_ind_y(new_center), cur_C(new_center), dim, k*up_num);
    if iter == iter_lim
        C = C+sparse(C_ind_x(cflag), C_ind_y(cflag), cur_C(cflag), dim, k*up_num);
    end

    iter = iter+1;
end

% make C a dim x k matrix
C_ind_x = repmat(1:dim,1,k);
C_ind_y = repmat(reshape(reshape(1:k*up_num,k,up_num)',1,k*up_num),G,1);
C=reshape(full(C(C_ind_x(:)+(C_ind_y(:)-1)*dim)),dim,k);
fprintf('\n');

% % work only if G = 2
% for ind=1:up_num
%     scatter(X(G*ind-1,:),X(G*ind,:),5,label(ind:up_num:(num-1)*up_num+ind))
%     hold on
%     scatter(C(G*ind-1,:),C(G*ind,:),100,'black','filled')
%     hold off
%     drawnow;
%     pause(1)
% end

end
