function varargout = AROWwOVR( C, beta, iternum, code, labels, codebook, mu, sigma )
% Adaptive Regularization Of Weight vectors
% Input
%   C               aggressiveness parametor (upper bound of tau)
%   beta            # neg. sample / # pos. sample
%   iternum         # of all iteration
%   code            dim * datanum training matrix
%   labels          datanum vector
%   codebook        dim * k matrix of PQ codebooks
%   mu              dim+1 * classnum initial weights (default: zeros)
%   sigma           dim+1 * classnum initial weights (default: zeros)
% Output
%   mu              dim+1 * classnum weights
%   sigma           dim+1 * classnum sigma
%
% Copyright (c) 2014, Yoshitaka Ushiku
% This software is released under the MIT License.
% http://opensource.org/licenses/mit-license.php

labels = online_learning.labvec2labmat(labels);
n = size(labels,1);

% evaluation setting
if nargin < 6 || isempty(codebook)
    isPQ = false;
    d = size(code,1);
else
    isPQ = true;
    d = size(codebook,1);
    G = d / size(code,1);
end

% training and testing
if nargin < 7 || isempty(mu)
    mu = zeros(d+1,n);
    sigma = ones(d+1,n);
end

pos_smpl_num = sum(labels,2)';
hingelosses = zeros(1,n);
is_averaging = false;
for iter = 1:iternum

    if iter==iternum && nargout == 3
        is_averaging = true;
        dmu = zeros(size(mu));
    end
    updates = zeros(1,n);
    neg_smpl_prob = pos_smpl_num * beta ./ (size(labels,2)-pos_smpl_num);

    for trnind = online_learning.randpermrs(size(labels,2))

        if sum(labels(:,trnind))==0; continue; end
        if isPQ
            cb_ind = double(reshape(code(:,trnind*ones(G,1))',d,1))*d + (1:d)';
            F = [codebook(cb_ind); 1];
        else
            F = [code(:,trnind); 1];
        end
        y = sign(labels(:,trnind)-0.5)';
        cur_train = y == 1 | (y == -1 & (rand(1,n) < neg_smpl_prob));
        hingelosses(cur_train) = 1-F'*mu(:,cur_train).*y(cur_train);
        cur_update = hingelosses > 0 & cur_train;

        F_2 = F.^2;
        updates = updates + cur_train;
        c = F_2' * sigma(:, cur_update) + 1/C;
        sigma_F = bsxfun(@times, sigma(:, cur_update), F * (y(cur_update).*hingelosses(cur_update)./c));

        mu(:,cur_update) = mu(:,cur_update) + sigma_F;
        sigma(:,cur_update) = 1./bsxfun(@plus,1./sigma(:,cur_update),F_2 * C);
        % averaging
        if is_averaging
            dmu(:,cur_update) = dmu(:,cur_update) + bsxfun(@times,updates(cur_update),sigma_F);
        end

    end
    
    fprintf('iter\t%d\t\tupdated %d \033[K\n',iter,size(labels,2));

end

if nargout == 3
    updates = updates + 1;
    avgmu = mu-bsxfun(@rdivide,dmu,updates);
    varargout{1} = avgmu;
    varargout{2} = mu;
    varargout{3} = sigma;
else
    varargout{1} = mu;
    varargout{2} = sigma;
end

end


