function varargout = CWuOVR( eta, iternum, code, labels, codebook, mu, sigma )
% Confidence Weighted
% Input
%   eta             probability of classifying correctly
%   iternum         # of all iteration
%   code            dim * datanum training matrix
%   labels          datanum vector
%   codebook        dim * k matrix of PQ codebooks
%   mu              dim+1 * classnum initial weights (default: zeros)
%   sigma           dim+1 * classnum initial weights (default: zeros)
% Outpute
%   avgmu           dim+1 * classnum averaged weights
%   mu              dim+1 * classnum weights
%   sigma           dim+1 * classnum sigma
%
% Copyright (c) 2014, Yoshitaka Ushiku
% This software is released under the MIT License.
% http://opensource.org/licenses/mit-license.php

phi = cdf('Normal',eta,0,1)^-1;

labels = online_learning.labvec2labmat(labels);
n = size(labels,1);

% evaluation setting
if nargin < 5 || isempty(codebook)
    isPQ = false;
    d = size(code,1);
else
    isPQ = true;
    d = size(codebook,1);
    G = d / size(code,1);
end

% training and testing
if nargin < 6 || isempty(mu)
    mu = zeros(d+1,n);
    sigma = ones(d+1,n);
end

is_averaging = false;
for iter = 1:iternum

    if iter==iternum && nargout == 3
        is_averaging = true;
        dmu = zeros(size(mu));
    end
    update = 0;

    for trnind = online_learning.randpermrs(size(labels,2))

        if sum(labels(:,trnind))==0; continue; end
        if isPQ
            cb_ind = double(reshape(code(:,trnind*ones(G,1))',d,1))*d + (1:d)';
            F = [codebook(cb_ind); 1];
        else
            F = [code(:,trnind); 1];
        end
        y = sign(labels(:,trnind)-0.5)';
        margin = F'*mu.*y;
        F_2 = F.^2;
        x_sigma_x = F_2' * sigma;
        cur_update = margin < phi * sqrt(x_sigma_x);

        update = update + 1;
        sub_alpha = -margin(cur_update) * (1 + phi^2/2);
        alpha = max(0,(sub_alpha + sqrt(margin(cur_update).^2*phi^4/4+x_sigma_x(cur_update)*phi^2*(1+phi^2))) ...
                    ./( x_sigma_x(cur_update) * (1 + phi^2) ));
        sigma_factor = -x_sigma_x(cur_update)+sqrt(x_sigma_x(cur_update).^2+4*x_sigma_x(cur_update)./(alpha*phi).^2);
        sigma_F = bsxfun(@times, bsxfun(@times,sigma(:,cur_update),F), alpha.*y(cur_update));

        mu(:,cur_update) = mu(:,cur_update) + sigma_F;
        sigma(:,cur_update) = 1./(1./sigma(:,cur_update) + F_2*(2./sigma_factor));

        if is_averaging
            dmu(:,cur_update) = dmu(:,cur_update) + update*sigma_F;
        end

    end
    
    fprintf('iter\t%d\t\tupdated %d \033[K\n',iter,size(labels,2));

end

if nargout == 3
    update = update + 1;
    avgmu = mu-dmu/update;
    varargout{1} = avgmu;
    varargout{2} = mu;
    varargout{3} = sigma;
else
    varargout{1} = mu;
    varargout{2} = sigma;
end

end
