function varargout = PAuOVR( C, iternum, code, labels, codebook, mu )
% Averaged Passive-Aggressive algorithm using OVR manner with all samples
% Input
%   C               aggressiveness parametor (upper bound of tau)
%   iternum         # of all iteration
%   code            dim * datanum training matrix
%   labels          datanum vector or classnum * datanum logical matrix
%   codebook        dim * k matrix of PQ codebooks
%                   if empty, usual PA (but OVR) is performed
%   mu              dim+1 * classnum initial weights (default: zeros)
% Output
%   avgmu           dim+1 * classnum averaged weights
%   mu              dim+1 * classnum weights
%
% Copyright (c) 2014, Yoshitaka Ushiku
% This software is released under the MIT License.
% http://opensource.org/licenses/mit-license.php

labels = online_learning.labvec2labmat(labels);
n = size(labels,1);

% evaluation setting
if nargin < 5 || isempty(codebook)
    isPQ = false;
    d = size(code,1);
else
    isPQ = true;
    d = size(codebook,1);
    G = d / size(code,1);
end

% training and testing
if nargin < 6 || isempty(mu)
    mu = zeros(d+1,n);
end

is_averaging = false;
for iter = 1:iternum

    if iter==iternum && nargout == 2
        is_averaging = true;
        dmu = zeros(size(mu));
    end
    update = 0;

    for trnind = online_learning.randpermrs(size(labels,2))

        if sum(labels(:,trnind))==0; continue; end
        if isPQ
            cb_ind = double(reshape(code(:,trnind*ones(G,1))',d,1))*d + (1:d)';
            F = [codebook(cb_ind); 1];
        else
            F = [code(:,trnind); 1];
        end
        y = sign(labels(:,trnind)-0.5)';
        hingelosses = 1-F'*mu.*y;
        cur_update = hingelosses > 0;

        update = update + 1;
        taus_F = F*(min(C,hingelosses(cur_update)/(F'*F)).*y(cur_update));

        mu(:,cur_update) = mu(:,cur_update)+taus_F;
        % averaging
        if is_averaging
            dmu(:,cur_update) = dmu(:,cur_update)+update*taus_F;
        end

    end
    
    fprintf('iter\t%d\t\tupdated %d \033[K\n',iter,update);

end

if nargout == 2
    update = update + 1;
    avgmu = mu-dmu/update;
    varargout{1} = avgmu;
    varargout{2} = mu;
else
    varargout{1} = mu;
end

end
